"""Base class for machine groups."""
from collections import defaultdict, namedtuple
from dataclasses import dataclass
from typing import Optional, Union, List
from abc import ABC, abstractmethod
import datetime
import time
import enum
import math

import logging

import inductiva
import inductiva.client
import inductiva.client.models
from inductiva import api, users, logs
from inductiva.commands.mpiconfig import MPIConfig
from inductiva.resources.utils import ProviderType
from inductiva.utils import format_utils

VCPUCount = namedtuple("VCPUCount", ["total", "per_machine"])


class ResourceType(enum.Enum):
    """Enum to represent the type of machine to be launched."""

    STANDARD = "standard"
    MPI = "mpi"


@dataclass(repr=False)
class BaseMachineGroup(ABC):
    """Base class to manage Google Cloud resources.

    Args:
        machine_type: The type of GC machine to launch. Ex: "e2-standard-4".
          Check https://cloud.google.com/compute/docs/machine-resource for
          more information about machine types.
        zone: The zone where the machines will be launched.
        provider: The cloud provider of the machine group.
        threads_per_core: The number of threads per core (1 or 2).
        data_disk_gb: The size of the disk for user data (in GB).
        max_idle_time: Time without executing any task, after which the
          resource will be terminated. Can be an exact timedelta or an int
            representing the number of minutes.
        auto_terminate_ts: Moment in which the resource will be
          automatically terminated.
        auto_terminate_minutes: Duration, in minutes, the MPICluster will be
                kept alive. After auto_terminate_minutes minutes the machine
                will be terminated. This time will start counting after calling
                this method.
        mpi_version: The version of MPI to be used on the machines.
        np: The number of processes to use for MPI commands.
        use_hwthread_cpus: Whether to use hyperthreading or not.

    :meta private:
    """
    # Constructor arguments
    machine_type: str
    zone: Optional[str] = "europe-west1-b"
    provider: Union[ProviderType, str] = "GCP"
    threads_per_core: int = 2
    data_disk_gb: int = 10
    max_idle_time: Union[datetime.timedelta, int] = 3
    auto_terminate_ts: Optional[datetime.datetime] = None
    auto_terminate_minutes: Optional[int] = None
    spot: bool = True

    mpi_version: str = "4.1.6"
    np: int = None
    use_hwthread_cpus: bool = True

    create_time = None
    num_machines = 0
    quota_usage = {}
    allow_auto_start = True

    # Internal attributes
    _free_space_threshold_gb = 5
    _id = None
    _name = None
    _started = False
    #Number of active machines at the time of
    #the request machine_groups.get()
    _active_machines = 0
    _custom_vm_image = None
    _estimated_cost = None
    _idle_seconds = None
    _cost_per_hour = {}
    _total_ram_gb = None
    _cpu_info = {}
    _gpu_info = {}

    QUOTAS_EXCEEDED_SLEEP_SECONDS = 60

    def __post_init__(self):
        """Validate inputs and initialize additional attributes after
        dataclass initialization."""
        provider = ProviderType(self.provider)
        self.provider = provider.value

        # Set the API configuration that carries the information from the client
        # to the backend.
        self._api = inductiva.client.ComputeApi(api.get_client())

        self._validate_inputs()

    def _validate_inputs(self):
        """Validate initialization inputs."""
        if not isinstance(self.data_disk_gb, int):
            raise ValueError("`data_disk_gb` must be an integer.")

        if not 10 <= self.data_disk_gb <= 65536:
            raise ValueError("`data_disk_gb` must be between 10 and 65536.")

        if self.auto_resize_disk_max_gb is not None:
            if not isinstance(self.auto_resize_disk_max_gb,
                              int) or self.auto_resize_disk_max_gb <= 0:
                raise ValueError(
                    "`auto_resize_disk_max_gb` must be a positive integer.")

            if self.auto_resize_disk_max_gb < self.data_disk_gb:
                raise ValueError(
                    "`auto_resize_disk_max_gb` must be greater than \
                    or equal to `data_disk_gb GB`.")

        if self.threads_per_core not in [1, 2]:
            raise ValueError("`threads_per_core` must be either 1 or 2.")

        if isinstance(self.max_idle_time, int):
            if self.max_idle_time <= 0:
                raise ValueError("`max_idle_time` must be positive.")
            self._max_idle_time = datetime.timedelta(minutes=self.max_idle_time)

        if self.auto_terminate_ts is not None:
            logging.warning("You are using `auto_terminate_ts`. This argument"
                            "will be deprecated in the future. Please use"
                            "`auto_terminate_minutes` instead.")

        if isinstance(self.auto_terminate_minutes, int):
            time_delta_minutes = datetime.timedelta(
                minutes=self.auto_terminate_minutes)
            self._auto_terminate_ts = datetime.datetime.now(
                tz=datetime.timezone.utc) + time_delta_minutes

    def has_gpu(self) -> bool:
        """Check if the machine group has a GPU."""
        if self._gpu_info is None:
            return False
        return self._gpu_info.gpu_count > 0

    def gpu_count(self) -> bool:
        """
        Returns the number of GPUs available in the resource.
        """
        if self._gpu_info is None:
            return 0
        return self._gpu_info.get("gpu_count", 0)

    def set_mpi_config(self,
                       mpi_version: str = "4.1.6",
                       np: Optional[int] = None,
                       use_hwthread_cpus: bool = True):
        """Set the MPI configuration for the cluster.
        Args:
            mpi_version: The version of MPI to be used on the machines.
            np: The number of processes to use for MPI commands.
            use_hwthread_cpus: Whether to use hyperthreading or not.
        """
        self.mpi_version = mpi_version
        self.use_hwthread_cpus = use_hwthread_cpus
        self.np = np or self.available_vcpus

    def get_mpi_config(self):
        """Get the MPI configuration for the cluster."""
        return MPIConfig(self.mpi_version,
                         np=self.np,
                         use_hwthread_cpus=self.use_hwthread_cpus)

    @property
    def id(self):
        return self._id

    @property
    @abstractmethod
    def n_vcpus(self):
        """Returns the number of vCPUs available in the resource.

        Returns a tuple with the total number of vCPUs and the number of vCPUs
        per machine. For a machine group with 2 machines, each with 4 vCPUs,
        this will return (8, 4).
        In a case of an Elastic machine group, the total number of vCPUs is
        the maximum number of vCPUs that can be used at the same time.
        """
        pass

    @property
    def available_vcpus(self):
        """Returns the maximum number of vCPUs that can be used on a task.

        On a machine group with 2 machines, each with 4 vCPUs, this will return
        4.
        On an elastic machine group, this will also return 4.
        On an MPI cluster this will return the total number of vcpus because
        we can run on the total number of vcpus.
        """
        return self.n_vcpus.per_machine

    def get_available_mpi_slots(
        self,
        use_hwthread: Optional[bool] = None,
    ) -> int:
        if use_hwthread is None:
            use_hwthread = self.use_hwthread_cpus

        n_cores = self.available_vcpus / self.threads_per_core

        threads_per_core = self.threads_per_core
        if not use_hwthread:
            threads_per_core = 1

        return int(n_cores * threads_per_core)

    @property
    def name(self):
        return self._name

    @property
    def started(self):
        return self._started

    @property
    def idle_time(self) -> datetime.timedelta:
        """
        Resource idle time in seconds.
        """
        return self._idle_seconds

    @property
    def total_ram_gb(self):
        return self._total_ram_gb

    @abstractmethod
    def short_name(self) -> str:
        pass

    @staticmethod
    def _timedelta_to_seconds(value: Union[datetime.timedelta, int]) -> float:
        """Converts a timedelta object to seconds."""
        if isinstance(value, int):
            return value * 60
        return value.total_seconds()

    @staticmethod
    def _seconds_to_timedelta(
            value: Optional[float] = None) -> Optional[datetime.timedelta]:
        """Converts seconds to a timedelta object."""
        return datetime.timedelta(seconds=float(value)) if value else None

    @staticmethod
    def _convert_auto_terminate_ts(
            timestamp: Optional[datetime.datetime]) -> Optional[str]:
        """Converts a datetime object to ISO format. Additionally, it checks
        if the datetime is in the future and if it is timezone aware."""
        if timestamp is not None:
            if (timestamp.tzinfo is None or
                    timestamp.tzinfo.utcoffset(timestamp) is None):
                raise ValueError("auto_terminate_ts must be timezone aware.")

            now_ts = datetime.datetime.now(timestamp.tzinfo)
            if timestamp < now_ts:
                raise ValueError("auto_terminate_ts must be in the future.")
            return timestamp.isoformat()

        return None

    def _dynamic_disk_resize_config(self):
        if self.auto_resize_disk_max_gb is None:
            return None

        return {
            "free_space_threshold_gb": self._free_space_threshold_gb,
            "max_disk_size_gb": self.auto_resize_disk_max_gb
        }

    @staticmethod
    def _iso_to_datetime(
            timestamp: Optional[str]) -> Optional[datetime.datetime]:
        """Converts an ISO format string back to a datetime object. It ensures
        the datetime is timezone aware."""
        if timestamp:
            dt = datetime.datetime.fromisoformat(str(timestamp))
            if dt.year == 9999:
                return None

            if dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None:
                raise ValueError("The datetime string must be timezone aware.")
            return dt
        return None

    def _update_attributes_from_response(
            self, resp: inductiva.client.models.VMGroupConfig):
        """Update machine group attributes with values from the API response."""
        self._id = resp.id
        self._name = resp.name
        self.quota_usage = resp.quota_usage
        # Lifecycle configuration parameters are updated with default values
        # from the API response if they were not provided by the user
        self.max_idle_time = self._seconds_to_timedelta(resp.max_idle_time)
        self._idle_seconds = self._seconds_to_timedelta(resp.idle_seconds)
        self.auto_terminate_ts = self._iso_to_datetime(resp.auto_terminate_ts)
        self._total_ram_gb = resp.total_ram_gb
        self._cost_per_hour = resp.cost_per_hour
        self._cpu_info = resp.cpu_info
        self._gpu_info = resp.gpu_info
        self.zone = resp.zone
        dynamic_disk_resize_config = resp.dynamic_disk_resize_config
        self.auto_resize_disk_max_gb = (
            dynamic_disk_resize_config.max_disk_size_gb
            if dynamic_disk_resize_config else None)

    def _register_machine_group(self, **kwargs):
        """Register machine group configuration in API.

        Returns:
            The unique ID and name identifying the machine on the API."""
        logging.info("■ Registering %s configurations:", self.short_name())

        instance_group_config = inductiva.client.models.RegisterVMGroupRequest(
            machine_type=self.machine_type,
            provider_id=self.provider,
            threads_per_core=self.threads_per_core,
            disk_size_gb=self.data_disk_gb,
            max_idle_time=self._timedelta_to_seconds(self.max_idle_time),
            auto_terminate_ts=self._convert_auto_terminate_ts(
                self.auto_terminate_ts),
            dynamic_disk_resize_config=self._dynamic_disk_resize_config(),
            custom_vm_image=self._custom_vm_image,
            zone=self.zone,
            **kwargs,
        )

        body = self._api.register_vm_group(
            register_vm_group_request=instance_group_config,)

        self._update_attributes_from_response(body)

        self._log_machine_group_info()

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}(name=\"{self.name}\")"

    def active_machines_to_str(self) -> str:
        """Return the number of machines currently running.
        """
        return f"{self._active_machines}/{self.num_machines}"

    @classmethod
    def from_api_response(cls, resp: inductiva.client.models.VMGroupConfig):
        """Creates a MachineGroup object from an API response."""

        # Do not call __init__ to prevent registration of the machine group
        machine_group = cls.__new__(cls)
        machine_group._api = inductiva.client.ComputeApi(api.get_client())
        machine_group.machine_type = resp.machine_type
        machine_group.data_disk_gb = resp.disk_size_gb
        machine_group.provider = resp.provider_id
        machine_group.create_time = resp.creation_timestamp
        machine_group.spot = bool(resp.spot)
        machine_group._started = bool(resp.started)
        machine_group.__dict__["machines"] = resp.machines
        machine_group.__dict__["_active_machines"] = int(resp.num_vms)

        machine_group._update_attributes_from_response(resp)

        machine_group.set_mpi_config()

        return machine_group

    def can_start_resource(self) -> bool:
        """Check if the resource can be started.

        This method checks if the resource can be started by checking
        the available quotas and resource usage.

        returns:
            bool: True if the resource can be started, False otherwise.
        """
        quotas = users.get_quotas()

        cost_in_use = quotas["max_price_hour"].in_use
        cost_max = math.inf if quotas[
            "max_price_hour"].max_allowed is None else quotas[
                "max_price_hour"].max_allowed
        estimated_cost = cost_in_use + self.estimate_cloud_cost(verbose=False)
        is_cost_ok = estimated_cost <= cost_max

        vcpu_in_use = quotas["max_vcpus"].in_use
        vcpu_max = (math.inf if quotas["max_vcpus"].max_allowed is None else
                    quotas["max_vcpus"].max_allowed)
        current_vcpu = self.n_vcpus.total
        estimated_vcpu_usage = vcpu_in_use + current_vcpu
        is_vcpu_ok = estimated_vcpu_usage <= vcpu_max

        machines_in_use = quotas["max_instances"].in_use
        machines_max = math.inf if quotas[
            "max_instances"].max_allowed is None else quotas[
                "max_instances"].max_allowed

        estimated_machine_usage = machines_in_use + self.num_machines
        is_instance_ok = estimated_machine_usage <= machines_max

        return is_cost_ok and is_vcpu_ok and is_instance_ok

    @logs.mute_logging()
    def start(self, wait_for_quotas: bool = False, verbose: bool = True):  # pylint: disable=unused-argument
        """Starts a machine group.

        Args:
            wait_for_quotas: If True, the method will wait for quotas to
              become available before starting the resource."""
        if self._started:
            logging.info("Attempting to start a machine group already started.")
            return

        if self.id is None or self.name is None:
            logging.info("Attempting to start an unregistered machine group. "
                         "Make sure you have called the constructor.")
            return

        logging.info(
            "Starting %s. This may take a few minutes.\n"
            "Note that stopping this local process will not interrupt "
            "the creation of the machine group. Please wait...", repr(self))
        start_time = time.time()

        if wait_for_quotas:
            if not self.can_start_resource():
                print("This machine will exceed the current quotas.\n"
                      "Will wait for quotas to become available.")
            while not self.can_start_resource():
                time.sleep(self.QUOTAS_EXCEEDED_SLEEP_SECONDS)

        self._api.start_vm_group(machine_group_id=self.id)
        creation_time = format_utils.seconds_formatter(time.time() - start_time)
        self._started = True
        quota_usage_table_str = self.quota_usage_table_str("used by resource")
        logging.info(
            "%s successfully started in %s.\n\n"
            "The machine group is using the following quotas:\n"
            "%s", self, creation_time, quota_usage_table_str)
        return True

    def terminate(self, verbose: bool = True):
        """Terminates a machine group."""
        if not self._started or self.id is None or self.name is None:
            logging.warning(
                "Attempting to terminate an unstarted machine group.")
            return

        try:
            self._api.delete_vm_group(machine_group_id=self.id)
            if verbose:
                logging.info("Successfully requested termination of %s.",
                             repr(self))
                if self.provider == ProviderType.GCP:
                    logging.info("Termination of the machine group "
                                 "freed the following quotas:")
                    logging.info(
                        self.quota_usage_table_str("freed by resource"))
            return True

        except inductiva.client.ApiException as api_exception:
            raise api_exception

    def _get_estimated_cost(self, spot: bool = True) -> float:
        """Returns estimate cost of a single machine in the group.

        This method is an overlay of the more general method, but
        it verifies if the cost has already been estimated and returns
        it immediately if it has.
        """
        if self.provider in (ProviderType.LOCAL,):
            return 0

        self._estimated_cost = inductiva.resources.estimate_machine_cost(
            self.machine_type,
            spot,
            self.zone,
        )

        return self._estimated_cost

    def quota_usage_table_str(self, resource_usage_header: str) -> str:
        quotas = users.get_quotas()
        table = defaultdict(list)
        emph_formatter = format_utils.get_ansi_formatter()

        header_formatters = [
            lambda x: emph_formatter(x.upper(), format_utils.Emphasis.BOLD)
        ]

        def _format_float(x):
            return f"{x:.3f}" if isinstance(x, float) else str(x)

        formatters = {
            column: [_format_float] for column in
            [resource_usage_header, "current usage", "max allowed"]
        }

        for name, value in self.quota_usage.items():
            quota = quotas.get(name)
            if not quota:
                max_allowed = "n/a"
                full_name = "n/a"
                in_use = "n/a"

            max_allowed = quota.max_allowed
            full_name = quota.label
            in_use = quota.in_use

            table[""].append(full_name)
            table[resource_usage_header].append(value)
            table["current usage"].append(in_use)
            table["max allowed"].append(max_allowed)

        table_str = format_utils.get_tabular_str(
            table,
            formatters=formatters,
            header_formatters=header_formatters,
        )

        return table_str

    def _log_estimated_spot_vm_savings(self) -> None:
        if self.provider in (ProviderType.LOCAL,):
            return

        spot_cost = self._get_estimated_cost(True)
        non_spot_cost = self._get_estimated_cost(False)
        spot_times_cheaper = round(non_spot_cost / spot_cost, 2)

        is_spot = getattr(self, "spot", True)

        if not is_spot:
            logging.info(
                "\t· The same machine group with spot machines would cost "
                "%.1fx less. Specify "
                "`spot=True` in the constructor to use spot machines.",
                spot_times_cheaper,
            )
        else:
            logging.info(
                "\t· You are spending %.1fx less "
                "by using spot machines.",
                spot_times_cheaper,
            )
        logging.info("")

    def _log_machine_group_info(self):
        """Logs the machine group info."""

        logging.info("\t· Name:                       %s", self.name)
        logging.info("\t· Provider:                   %s", self.provider)
        logging.info("\t· Machine Type:               %s", self.machine_type)
        logging.info("\t· Data disk size:             %s GB", self.data_disk_gb)

        if self.auto_resize_disk_max_gb:
            logging.info("\t· Auto resize disk max size:  %s GB",
                         self.auto_resize_disk_max_gb)

        logging.info("\t· Total memory (RAM):         %s GB", self.total_ram_gb)

        # Log max idle time
        value_str = format_utils.timedelta_formatter(
            self.max_idle_time) if self.max_idle_time is not None else "N/A"
        logging.info("\t· Maximum idle time:          %s", value_str)

        # Log auto terminate timestamp
        value_str = self.auto_terminate_ts.strftime(
            "%Y/%m/%d %H:%M:%S"
        ) if self.auto_terminate_ts is not None else "N/A"
        logging.info("\t· Auto terminate timestamp:   %s", value_str)

    def estimate_cloud_cost(self, verbose: bool = True):
        """Estimates a cost per hour of min and max machines in US dollars.

        these are the estimted costs of having minimum and the
        maximum number of machines up in the cloud. The final cost will vary
        depending on the total usage of the machines."""

        min_cost_per_hour = self._cost_per_hour.min
        max_cost_per_hour = self._cost_per_hour.max

        if not verbose:
            return max_cost_per_hour

        if min_cost_per_hour == max_cost_per_hour:
            logging.info(
                "\t· Estimated cloud cost of machine group: %.3f $/h",
                max_cost_per_hour,
            )
        else:
            min_reason = self._cost_per_hour.min_reason.rstrip(".")
            max_reason = self._cost_per_hour.max_reason.rstrip(".")

            logging.info("\t· Estimated cloud cost of machine group:")
            logging.info("\t\t· Minimum: %.3f $/h (%s)", min_cost_per_hour,
                         min_reason)
            logging.info("\t\t· Maximum: %.3f $/h (%s)", max_cost_per_hour,
                         max_reason)

        self._log_estimated_spot_vm_savings()

        return max_cost_per_hour


@dataclass(repr=False)
class MachineGroup(BaseMachineGroup):
    """Create a MachineGroup object.

    A machine group is a collection of homogenous machines with given the
    configurations that are launched in Google Cloud.
    Note: The machine group will be available only after calling 'start' method.
    The billing will start only after the machines are started.

    Args:
        machine_type: The type of GC machine to launch. Ex: "e2-standard-4".
          Check https://cloud.google.com/compute/docs/machine-resource for
          information about machine types.
        zone: The zone where the machines will be launched.
        provider: The cloud provider of the machine group.
        threads_per_core: The number of threads per core (1 or 2).
        data_disk_gb: The size of the disk for user data (in GB).
        max_idle_time: Time without executing any task, after which the
          resource will be terminated.
        auto_terminate_ts: Moment in which the resource will be
          automatically terminated.
        auto_terminate_minutes: Duration, in minutes, the MPICluster will be
                kept alive. After auto_terminate_minutes minutes the machine
                will be terminated. This time will start counting after calling
                this method.
        auto_resize_disk_max_gb: The maximum size in GB that the hard disk
            of the cloud VM can reach. If set, the disk will be
            automatically resized, during the execution of a task, when the
            free space falls below a certain threshold. This mechanism helps
            prevent "out of space" errors, that can occur when a task
            generates a quantity of output files that exceeds the size of
            the local storage. Increasing disk size during task execution
            increases the cost of local storage associated with the VM,
            therefore the user must set an upper limit to the disk size, to
            prevent uncontrolled costs. Once that limit is reached, the disk
            is no longer automatically resized, and if the task continues to
            output files, it will fail.
        num_machines: The number of virtual machines to launch.
        spot: Whether to use spot machines.
    """
    # Constructor arguments
    auto_resize_disk_max_gb: Optional[int] = None
    num_machines: int = 1

    # Internal attributes
    _is_elastic = False

    def __post_init__(self):
        """Validate inputs and initialize additional attributes after
        dataclass initialization."""
        super().__post_init__()

        self._register_machine_group(num_vms=self.num_machines,
                                     spot=self.spot,
                                     is_elastic=self._is_elastic)
        self.np = self.np or self.available_vcpus

    def _validate_inputs(self):
        super()._validate_inputs()
        if self.num_machines < 1:
            raise ValueError(
                "`num_machines` should be a number greater than 0.")

    @property
    def n_vcpus(self):
        return VCPUCount(self._cpu_info.cpu_cores_logical * self.num_machines,
                         self._cpu_info.cpu_cores_logical)

    def short_name(self) -> str:
        return "MachineGroup"

    @classmethod
    def from_api_response(cls, resp: inductiva.client.models.VMGroupConfig):
        machine_group = super().from_api_response(resp)
        machine_group.num_machines = int(resp.max_vms)
        return machine_group

    def __str__(self):
        return f"Machine Group {self.name} with {self.machine_type} machines"

    def _log_machine_group_info(self):
        super()._log_machine_group_info()
        logging.info("\t· Number of machines:         %s", self.num_machines)
        logging.info("\t· Spot:                       %s", self.spot)
        self.estimate_cloud_cost()


@dataclass(repr=False)
class ElasticMachineGroup(BaseMachineGroup):
    """Create an ElasticMachineGroup object.

    An ElasticMachineGroup is a set of identical machines that can
    automatically scale based on CPU load. The group starts with a
    minimum number of machines and adjusts its size as needed scaling
    to the maximum number of machines, ensuring both optimal performance
    and cost efficiency.
    Note: The machine group becomes active after calling the 'start' method,
    and billing commences once the machines are initiated.

    Args:
        machine_type: The type of GC machine to launch. Ex: "e2-standard-4".
            Check https://cloud.google.com/compute/docs/machine-resource for
        more information about machine types.
        zone: The zone where the machines will be launched.
        provider: The cloud provider of the machine group.
        threads_per_core: The number of threads per core (1 or 2).
        data_disk_gb: The size of the disk for user data (in GB).
        max_idle_time: Time without executing any task, after which the
            resource will be terminated.
        auto_terminate_ts: Moment in which the resource will be
            automatically terminated.
        auto_resize_disk_max_gb: The maximum size in GB that the hard disk
            of the cloud VM can reach. If set, the disk will be
            automatically resized, during the execution of a task, when the
            free space falls below a certain threshold. This mechanism helps
            prevent "out of space" errors, that can occur when a task
            generates a quantity of output files that exceeds the size of
            the local storage. Increasing disk size during task execution
            increases the cost of local storage associated with the VM,
            therefore the user must set an upper limit to the disk size, to
            prevent uncontrolled costs. Once that limit is reached, the disk
            is no longer automatically resized, and if the task continues to
            output files, it will fail.
        auto_terminate_minutes: Duration, in minutes, the MPICluster will be
                kept alive. After auto_terminate_minutes minutes the machine
                will be terminated. This time will start counting after calling
                this method.
        min_machines: The minimum number of available machines. This is
            a quantity of machines that will be started initially and the
            minimum available machines, even in cases of low CPU load.
        max_machines: The maximum number of machines a machine group
            can scale up to.
        spot: Whether to use spot machines.
    """
    # Constructor arguments
    auto_resize_disk_max_gb: Optional[int] = None
    min_machines: int = 1
    max_machines: int = 2

    # Internal attributes
    _is_elastic = True

    def __post_init__(self):
        """Validate inputs and initialize additional attributes after
        dataclass initialization."""
        super().__post_init__()

        self._active_machines = self.min_machines

        self._register_machine_group(min_vms=self.min_machines,
                                     max_vms=self.max_machines,
                                     is_elastic=self._is_elastic,
                                     num_vms=self._active_machines,
                                     spot=self.spot)
        self.np = self.np or self.available_vcpus

    def _validate_inputs(self):
        super()._validate_inputs()
        if self.min_machines < 0:
            raise ValueError(
                "`min_machines` should be a number equal or greater than 0.")

        if self.min_machines >= self.max_machines:
            raise ValueError("`max_machines` should be greater "
                             "than `min_machines`.")

    @property
    def n_vcpus(self):
        return VCPUCount(self._cpu_info.cpu_cores_logical * self.max_machines,
                         self._cpu_info.cpu_cores_logical)

    def short_name(self) -> str:
        return "ElasticMachineGroup"

    @classmethod
    def from_api_response(cls, resp: dict):
        machine_group = super().from_api_response(resp)
        machine_group.max_machines = int(resp.max_vms)
        machine_group.min_machines = int(resp.min_vms)
        return machine_group

    def active_machines_to_str(self) -> str:
        """Returns a string representation of the
        number of machines currently running.
        """
        return f"{self._active_machines}/{self.max_machines} (max)"

    def __str__(self):
        return f"Elastic Machine Group {self.name} with {self.machine_type} " \
             "machines"

    def _log_machine_group_info(self):
        super()._log_machine_group_info()
        logging.info("\t· Maximum number of machines: %s", self.max_machines)
        logging.info("\t· Minimum number of machines: %s", self.min_machines)
        logging.info("\t· Spot:                       %s", self.spot)
        self.estimate_cloud_cost()


@dataclass(repr=False)
class MPICluster(BaseMachineGroup):
    """Create a MPICluster object.

   A MPI cluster is a collection of homogenous machines all working together on
    a common task given the configurations that are launched in Google Cloud.
    Note: The cluster will be available only after calling 'start' method.
    The billing will start only after the machines are started.

    Args:
        machine_type: The type of GC machine to launch. Ex: "e2-standard-4".
            Check https://cloud.google.com/compute/docs/machine-resource for
            information about machine types.
        zone: The zone where the machines will be launched.
        provider: The cloud provider of the machine group.
        threads_per_core: The number of threads per core (1 or 2).
        data_disk_gb: The size of the disk for user data (in GB).
        max_idle_time: Time without executing any task, after which the
            resource will be terminated.
        auto_terminate_minutes: Duration, in minutes, the MPICluster will be
                kept alive. After auto_terminate_minutes minutes the machine
                will be terminated. This time will start counting after calling
                this method.
        auto_terminate_ts: Moment in which the resource will be
            automatically terminated.
        num_machines: The number of virtual machines to launch.
    """
    # Constructor arguments
    num_machines: int = 2

    # Internal attributes
    auto_resize_disk_max_gb = None
    _type = ResourceType.MPI.value
    _is_elastic = False

    def __post_init__(self):
        """Validate inputs and initialize additional attributes after
        dataclass initialization."""
        super().__post_init__()

        self._register_machine_group(num_vms=self.num_machines,
                                     is_elastic=self._is_elastic,
                                     spot=self.spot,
                                     type=self._type)
        self.np = self.np or self.available_vcpus

    def _validate_inputs(self):
        super()._validate_inputs()
        if self.num_machines < 1:
            raise ValueError(
                "`num_machines` should be a number greater than 0.")

    @property
    def n_vcpus(self):
        return VCPUCount(self._cpu_info.cpu_cores_logical * self.num_machines,
                         self._cpu_info.cpu_cores_logical)

    def gpu_count(self) -> bool:
        """
        Returns the number of GPUs available in the resource.
        """
        return super().gpu_count() * self.num_machines

    @property
    def available_vcpus(self):
        """Returns the number of vCPUs available to the resource.

        For a mpi cluster with 2 machines, each with 4 vCPUs, this will
        return 8.
        """

        return self.n_vcpus.total

    def short_name(self) -> str:
        return "MPICluster"

    @classmethod
    def from_api_response(cls, resp: dict):
        machine_group = super().from_api_response(resp)
        machine_group.num_machines = int(resp.max_vms)
        return machine_group

    def __str__(self):
        return f"MPI Cluster {self.name} with {self.machine_type} " \
               f"x{self.num_machines} machines"

    def _log_machine_group_info(self):
        super()._log_machine_group_info()
        logging.info("\t· Number of machines:       %s", self.num_machines)
        logging.info("\t· Spot:                     %s", self.spot)
        self.estimate_cloud_cost()

    def _log_estimated_spot_vm_savings(self) -> None:
        return


def _fetch_machine_groups_from_api(
) -> List[inductiva.client.models.VMGroupConfig]:
    """Get all active machine groups of a user from the API."""
    try:
        api_compute = inductiva.client.ComputeApi(inductiva.api.get_client())
        return api_compute.list_active_user_instance_groups()

    except inductiva.client.ApiException as api_exception:
        raise api_exception


def _get_machine_group_class(
    machine_type: inductiva.client.models.MachineGroupType,
    is_elastic: bool,
):
    """Returns the class of the machine group"""
    if is_elastic:
        mg_class = ElasticMachineGroup
    elif machine_type == inductiva.client.models.MachineGroupType.STANDARD:
        mg_class = MachineGroup
    elif machine_type == inductiva.client.models.MachineGroupType.MPI:
        mg_class = MPICluster
    else:
        raise ValueError("Unknown resource configuration.")
    return mg_class


def get_by_name(machine_name: str):
    """Returns the machine group corresponding to `machine_name`."""
    try:
        api_compute = inductiva.client.ComputeApi(inductiva.api.get_client())
        response = api_compute.get_vm_group_by_name(name=machine_name)
        mg_class = _get_machine_group_class(response.type, response.is_elastic)
        return mg_class.from_api_response(response)
    except inductiva.client.ApiException as api_exception:
        raise api_exception


def get():
    """Returns a list of 'Resource' objects."""

    # Retrive the active resource names
    machine_groups = _fetch_machine_groups_from_api()
    machine_group_list = []

    for mg in machine_groups:
        mg_class = _get_machine_group_class(mg.type, mg.is_elastic)
        machine_group_list.append(mg_class.from_api_response(mg))

    return machine_group_list



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, StrictFloat, StrictInt
from typing import Any, ClassVar, Dict, List, Optional, Union
from inductiva.client.models.currency_code import CurrencyCode
from inductiva.client.models.top_up_type import TopUpType
from typing import Optional, Set
from typing_extensions import Self


class Transaction(BaseModel):
    """
    Transaction
    """

  # noqa: E501
    amount: Union[StrictFloat, StrictInt]
    time: datetime
    currency: Optional[CurrencyCode] = None
    fee: Optional[Union[StrictFloat, StrictInt]] = None
    fee_percentage: Optional[Union[StrictFloat, StrictInt]] = None
    total: Optional[Union[StrictFloat, StrictInt]] = None
    top_up_type: TopUpType
    __properties: ClassVar[List[str]] = [
        "amount", "time", "currency", "fee", "fee_percentage", "total",
        "top_up_type"
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of Transaction from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # set to None if fee (nullable) is None
        # and model_fields_set contains the field
        if self.fee is None and "fee" in self.model_fields_set:
            _dict['fee'] = None

        # set to None if fee_percentage (nullable) is None
        # and model_fields_set contains the field
        if self.fee_percentage is None and "fee_percentage" in self.model_fields_set:
            _dict['fee_percentage'] = None

        # set to None if total (nullable) is None
        # and model_fields_set contains the field
        if self.total is None and "total" in self.model_fields_set:
            _dict['total'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of Transaction from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "amount": obj.get("amount"),
            "time": obj.get("time"),
            "currency": obj.get("currency"),
            "fee": obj.get("fee"),
            "fee_percentage": obj.get("fee_percentage"),
            "total": obj.get("total"),
            "top_up_type": obj.get("top_up_type")
        })
        return _obj



"""CLI commands to get costs of computational resources."""

import argparse

from inductiva import resources
from inductiva.utils import format_utils


def estimate_machine_cost(args):
    """Estimate the cost of a certain machine type."""
    machine_type = args.machine_type
    spot = args.spot
    num_machines = args.num_machines
    zone = args.zone

    cost = resources.estimate_machine_cost(
        machine_type=machine_type,
        spot=spot,
        zone=zone,
    )

    total_cost = cost * num_machines
    cost_str = format_utils.currency_formatter(cost)
    total_cost_str = format_utils.currency_formatter(total_cost)
    if num_machines == 1:
        print(f"Estimated cost of machine: {cost_str}/h.")
    else:
        print("Estimated total cost (per machine)"
              f": {total_cost_str} ({cost_str})/h.")


def register(parser):
    """Register the cost estimates commands."""

    subparser = parser.add_parser(
        "cost",
        help="Estimate cost of a machine in the cloud.",
        formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = (
        "The `inductiva cost` command provides a utility"
        " for estimating the cost\n"
        "of a machine in the cloud. It allows you to specify"
        " the type of machine\n"
        "and the number of machines, and it can calculate the"
        " cost for spot instances.\n\n")
    subparser.add_argument("machine_type",
                           type=str,
                           help="Type of machine to launch.")
    subparser.add_argument("--spot",
                           default=False,
                           action="store_true",
                           help="Type of machine to launch.")

    subparser.add_argument("-n",
                           "--num_machines",
                           default=1,
                           type=int,
                           help="Number of machines to launch.")

    subparser.add_argument("--zone",
                           type=str,
                           default="europe-west1-b",
                           help="Zone where the machines will be launched.")

    subparser.set_defaults(func=estimate_machine_cost)



"""Download the outputs of a task by ID via CLI."""
import argparse

import inductiva


def _download(download_func, filenames):
    download_func(filenames=filenames)


def download(args):
    """Download the outputs of a task by ID."""
    task_ids = args.task_id
    filenames = args.filenames
    download_dir = args.dir

    if not args.output and not args.input:
        args.output = True

    if download_dir is not None:
        inductiva.set_output_dir(download_dir)

    if download_dir == "":
        print("`dir` must not be an empty string.")
        return 1

    if not task_ids:
        print("No ID(s) specified.\n"
              "> Use `inductiva tasks download -h` for help.")
        return 1

    task_ids = set(task_ids)

    for task_id in task_ids:
        task = inductiva.tasks.Task(task_id)
        if args.output:
            _download(task.download_outputs, filenames)
        if args.input:
            _download(task.download_inputs, filenames)
    return 0


def register(parser):
    """Register the download task command."""

    subparser = parser.add_parser("download",
                                  help="Download tasks.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = (
        "Download the input/output files of tasks with the given ID(s).\n"
        "All files are downloaded unless the --filenames list is given,\n"
        "in which case only the indicated files for each task are downloaded.\n"
        "The name of the input/output folder can be configured through the\n"
        "--dir option. Files are downloaded to a subdirectory named after the\n"
        "ID of the corresponding task. The output files are downloaded when\n"
        "the --output/-o option is passed. Similarly, the input files are\n"
        "downloaded when the --input/-i option is passed. If neither option \n"
        "is specified, the output files are downloaded by default.")

    subparser.add_argument("task_id",
                           type=str,
                           help="ID(s) of the task(s) to download.",
                           nargs="*")
    subparser.add_argument("--filenames",
                           type=str,
                           help="Names of the files to download.",
                           nargs="*")
    subparser.add_argument("--dir",
                           type=str,
                           help="Path of where to download the task"
                           "input/output files.")
    subparser.add_argument("--input",
                           "-i",
                           action="store_true",
                           help="Option to download input files.")
    subparser.add_argument("--output",
                           "-o",
                           action="store_true",
                           help="Option to download output files (default).")

    subparser.set_defaults(func=download)



"""Remove a container file in remote storage."""

import argparse
import sys
from inductiva import storage
from inductiva.utils.input_functions import user_confirmation_prompt


def rm_container(args):
    default_folder = "my-containers"
    folder = args.folder or default_folder
    container_name = args.name

    if not container_name:
        print("Error: Container name is required.", file=sys.stderr)
        return 1

    container_path = f"{folder}/{container_name}"

    # Check if container exists
    try:
        contents = storage.listdir(folder, max_results=100, print_results=False)
    except Exception as e:  # pylint: disable=broad-except
        print(f"Error accessing remote folder '{folder}': {e}", file=sys.stderr)
        return 1

    if not any(
            content["content_name"] == container_name for content in contents):
        print(f"Error: '{container_name}' not found in folder '{folder}'.",
              file=sys.stderr)
        return 1

    # Ask for confirmation if not using --yes
    if not args.confirm:
        confirm = user_confirmation_prompt(
            [container_name],  # items
            "You are about to delete all containers.",
            f"You are about to delete {len([container_name])} containers.",
            "You are about to delete the following container:",
            is_all=False)

        if not confirm:
            print("Operation cancelled.")
            return 0

    try:
        storage.remove(container_path)
        print(f"✅ Container '{container_name}' removed.")
        return 0
    except Exception as e:  # pylint: disable=broad-except
        print(f"Error removing container: {e}", file=sys.stderr)
        return 1


def register(parser):
    """Register the remove-container command."""
    subparser = parser.add_parser(
        "remove",
        aliases=["rm"],
        help="Remove a container file from remote storage.",
        formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = (
        "Removes a specific container file from a remote storage folder.\n"
        "If no folder is specified, defaults to 'my-containers'.\n"
        "Use the `--yes` flag to skip confirmation prompts.")

    subparser.add_argument(
        "folder",
        nargs="?",
        type=str,
        help=
        "Optional folder path in remote storage. Defaults to 'my-containers'.")

    subparser.add_argument("-n",
                           "--name",
                           required=True,
                           type=str,
                           help="Name of the container to remove.")

    subparser.add_argument(
        "-y",
        "--yes",
        action="store_true",
        dest="confirm",
        default=False,
        help="Skip confirmation prompt and delete immediately.")

    subparser.set_defaults(func=rm_container)



"""Register CLI commands for logs."""
import os
import argparse

from inductiva import constants
from inductiva._cli import loader, utils


def register(root_parser):

    parser = root_parser.add_parser(
        "resources",
        help="Computational resource management utilities.",
        formatter_class=argparse.RawTextHelpFormatter)
    parser.description = (
        "Computational resource management utilities.\n\n"
        "The `inductiva resources` command provides utilities "
        "for managing computational resources.\nIt allows you "
        "to show estimated costs of resources, "
        "show available machine types, list current resources\n"
        " being used, and terminate resources.\n")

    utils.show_help_msg(parser)

    subparsers = parser.add_subparsers(title="available subcomands")
    loader.load_commands(subparsers,
                         os.path.dirname(__file__),
                         package=__name__,
                         ignores_prefix=constants.LOADER_IGNORE_PREFIX,
                         hides_prefix=constants.LOADER_HIDE_PREFIX)



# pylint: disable=missing-module-docstring
from . import files
from . import data
from .files import download_from_url
from .authentication import (
    get_stored_api_key,
    is_valid_token,
    remove_stored_api_key,
    set_stored_api_key,
)


def strtobool(val: str) -> bool:
    """Convert a string representation of truth to a boolean."""
    if isinstance(val, bool):
        return val

    if isinstance(val, str):
        val = val.lower()
        if val in ("y", "yes", "t", "true", "on", "1"):
            return True
        elif val in ("n", "no", "f", "false", "off", "0"):
            return False

    raise ValueError(f"Invalid truth value: {val}")



"""OpenTelemac module of the API."""
from typing import Optional, Union

from inductiva import simulators, types


class OpenTelemac(simulators.Simulator):
    """Class to invoke a generic OpenTelemac simulation on the API."""

    _default_opentelemac_device = "cpu"

    def __init__(
        self,
        /,
        version: Optional[str] = None,
        use_dev: bool = False,
    ):
        super().__init__(version, use_dev, self._default_opentelemac_device)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "opentelemac"

    def run(self,
            input_dir: Optional[str],
            commands: types.Commands,
            *,
            on: types.ComputationalResources,
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[Union[str, list[str]]] = None,
            project: Optional[str] = None,
            time_to_live: Optional[str] = None,
            **kwargs):

        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           project=project,
                           time_to_live=time_to_live,
                           **kwargs)



"""MPIConfig class"""


class MPIConfig:
    """MPIConfig class"""

    def __init__(self, version: str, **kwargs):
        """
        Args:
          version: The version of the MPI library to use.
            At the moment, only 1.10.7 and 4.1.6 are supported.
          **kwargs: Additional parameters or flags that Openmpi uses.
            parameters that have "-" should be typed with "_" instead.
            Example: np=2, use_hwthread_cpus=True
        """
        self.version = version
        self.options = kwargs

    def to_dict(self):
        """Convert this MPIConfig to a dictionary representation.
        This method will replace the "_" with "-".
        """
        return {
            "version": self.version,
            "options": {
                k.replace("_", "-"): v for k, v in self.options.items()
            }
        }



"""SCHISM module of the API."""
from typing import Optional, Union

from inductiva import simulators, tasks, types
from inductiva.commands.commands import Command
from inductiva.commands.mpiconfig import MPIConfig


@simulators.simulator.mpi_enabled
class SCHISM(simulators.Simulator):
    """Class to invoke a generic SCHISM simulation on the API."""

    def __init__(self, /, version: Optional[str] = None, use_dev: bool = False):
        """Initialize the SCHISM simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
        """
        super().__init__(version=version, use_dev=use_dev)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "schism"

    def run(self,
            input_dir: Optional[str],
            *,
            on: types.ComputationalResources,
            num_scribes: int = 1,
            storage_dir: Optional[str] = "",
            use_hwthread: bool = True,
            n_vcpus: Optional[int] = None,
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[Union[str, list[str]]] = None,
            project: Optional[str] = None,
            time_to_live: Optional[str] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.
        
        Args:
            input_dir: Path to the directory of the simulation input files.
            on: The computational resource to launch the simulation on.
            num_scribes: The num_scribes as per the simulator documentation.
            https://schism-dev.github.io/schism/master/getting-started/running-model.html
            storage_dir: Directory for storing simulation results.
            use_hwthread: If specified Open MPI will attempt to discover the
                number of hardware threads on the node, and use that as the
                number of slots available.
            n_vcpus: Number of virtual cpus
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
            project: Name of the project to which the task will be
                assigned. If None, the task will be assigned to
                the default project. If the project does not exist, it will be
                created.
            time_to_live: Maximum allowed runtime for the task, specified as a
                string duration. Supports common time duration formats such as
                "10m", "2 hours", "1h30m", or "90s". The task will be
                automatically terminated if it exceeds this duration after
                starting.
        """
        mpi_kwargs = {}
        mpi_kwargs["use_hwthread_cpus"] = use_hwthread
        if n_vcpus is not None:
            mpi_kwargs["np"] = n_vcpus

        mpi_config = MPIConfig(version="4.1.6", **mpi_kwargs)

        commands = [
            "mkdir -p outputs",
            Command(f"/schism/build/bin/pschism {num_scribes}",
                    mpi_config=mpi_config)
        ]

        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           project=project,
                           time_to_live=time_to_live,
                           **kwargs)



"""Register CLI commands for storage buckets."""
import argparse
import os

from inductiva import constants
from inductiva._cli import loader, utils


def register(root_parser):
    parser = root_parser.add_parser(
        "bucket",
        help="Storage bucket management utilities.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.description = (
        "Storage bucket management utilities.\n\n"
        "The `inductiva storage bucket` command provides tools for "
        "managing with cloud storage buckets.\n"
        "Use it to manage cloud storage buckets linked to your account.")

    utils.show_help_msg(parser)

    subparsers = parser.add_subparsers(title="available subcomands")
    loader.load_commands(subparsers,
                         os.path.dirname(__file__),
                         package=__name__,
                         ignores_prefix=constants.LOADER_IGNORE_PREFIX,
                         hides_prefix=constants.LOADER_HIDE_PREFIX)



"""AmrWind module of the API for numerical simulations of fluid flows."""
from typing import Literal, Optional, Union

from inductiva import simulators, tasks, types
from inductiva.commands import Command, MPIConfig


@simulators.simulator.mpi_enabled
class AmrWind(simulators.Simulator):
    """Class to invoke a generic AmrWind simulation on the API.
    """

    def __init__(self,
                 /,
                 version: Optional[str] = None,
                 use_dev: bool = False,
                 device: Literal["auto", "cpu", "gpu"] = "auto"):
        """Initialize the AmrWind simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
            device (str): The device to use for the simulation. If "auto",
                the device will be selected automatically. If "cpu", the
                simulation will run on the CPU. If "gpu", the simulation
                will run on the GPU.
        """
        super().__init__(version=version, use_dev=use_dev, device=device)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "amrwind"

    @property
    def name(self):
        """Get the name of the this simulator."""
        return "AMR-Wind"

    def run(self,
            input_dir: Optional[str],
            *,
            sim_config_filename: str,
            on: types.ComputationalResources,
            use_hwthread: bool = True,
            n_vcpus: Optional[int] = None,
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[Union[str, list[str]]] = None,
            project: Optional[str] = None,
            time_to_live: Optional[str] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            sim_config_filename: Name of the simulation configuration file.
            on: The computational resource to launch the simulation on.
            use_hwthread: If specified Open MPI will attempt to discover the
                number of hardware threads on the node, and use that as the
                number of slots available.
            n_vcpus: Number of vCPUs to use in the simulation. If not provided
                (default), all vCPUs will be used.
            storage_dir: Path to the directory where the simulation results
                will be stored.
            resubmit_on_preemption: Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
            project: Name of the project to which the task will be
                assigned. If None, the task will be assigned to
                the default project. If the project does not exist, it will be
                created.
            time_to_live: Maximum allowed runtime for the task, specified as a
                string duration. Supports common time duration formats such as
                "10m", "2 hours", "1h30m", or "90s". The task will be
                automatically terminated if it exceeds this duration after
                starting.
            **kwargs: Keyword arguments to be passed to the base class.
        """

        self._input_files_exist(input_dir=input_dir,
                                remote_assets=remote_assets,
                                sim_config_filename=sim_config_filename)

        mpi_kwargs = {}
        mpi_kwargs["use_hwthread_cpus"] = use_hwthread
        if n_vcpus is not None:
            mpi_kwargs["np"] = n_vcpus

        mpi_config = MPIConfig(version="4.1.6", **mpi_kwargs)
        commands = [
            Command(f"amr_wind {sim_config_filename}", mpi_config=mpi_config)
        ]

        return super().run(input_dir,
                           on=on,
                           storage_dir=storage_dir,
                           commands=commands,
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           project=project,
                           time_to_live=time_to_live,
                           **kwargs)



"""Download the project tasks files via CLI."""
import concurrent.futures as cf
import argparse
from typing import List

import inductiva
from inductiva.tasks.task import Task


def download_wrapper(task: Task,
                     output_dir: str = None,
                     filenames: List[str] = None):
    """Download the project tasks files.
    This method is a wrapper around the download_outputs method of the Task
    class. We do this to be able to set the outputdir for each worker.
    """
    if output_dir:
        inductiva.set_output_dir(output_dir)
    task.download_outputs(filenames)


def download_projects(args):
    """Download the project tasks files.

    This function downloads the files of the tasks of a project.
    Args:
        args.project_name (str): Name of the project to download.
        args.output_dir (str): Directory to save the downloaded files 
            default(inductiva-output).
        args.files (list): List of files to download.
        args.std (bool): Flag to download the standard output and error files.
    The downloads are done in parallel using a ThreadPoolExecutor with a maximum
    of 10 workers.
    """

    project_name = args.project_name
    output_dir = args.output_dir
    files = args.files
    std = args.std

    if std:
        files = ["stdout.txt", "stderr.txt"]

    project = inductiva.projects.Project(project_name)
    project_tasks = project.get_tasks()

    futures = []

    with cf.ThreadPoolExecutor(max_workers=10) as executor:
        for task in project_tasks:
            future = executor.submit(download_wrapper,
                                     task=task,
                                     output_dir=output_dir,
                                     filenames=files)
            futures.append(future)

    cf.wait(futures)
    print("Downloads completed.")


def register(parser):
    """Register the projects download command."""

    subparser = parser.add_parser(
        "download",
        help="Downloads the tasks files of a project.",
        formatter_class=argparse.RawTextHelpFormatter)

    subparser.add_argument("project_name",
                           type=str,
                           help="Name of the project to download.")

    subparser.description = (
        "The `inductiva projects download` command provides "
        "a way to download the tasks files of your projects.\n"
        "You can specify the files to download with the --files flag \n"
        "or download the standard output and error files with the --std flag.\n"
        "If no files are specified, all files are downloaded.\n")

    subparser.add_argument("--output-dir",
                           type=str,
                           help="Directory to save the downloaded files.")

    group = subparser.add_mutually_exclusive_group()
    group.add_argument("--files",
                       nargs="+",
                       type=str,
                       help="Downloads the specified files from every task in "
                       "the project.")
    group.add_argument("--std",
                       action="store_true",
                       help="Downloads the standard output and error files.")

    subparser.set_defaults(func=download_projects)



"""Type definitions."""
from typing import Union, List
from typing_extensions import TypeAlias

ComputationalResources: TypeAlias = Union["resources.MachineGroup",
                                          "resources.ElasticMachineGroup",
                                          "resources.MPICluster"]

Command: TypeAlias = Union[str, "inductiva.commands.Command"]
Commands: TypeAlias = List[Command]



"""Inductiva template manager class to manipulate files and render templates."""
import pathlib
import shutil
from . import helpers

import jinja2


class TemplateManager:
    """Template manager for rendering files and directories.

    Template files contains variables that are replaced by
    values provided as keyword arguments to the render methods.
    The underlying template engine is the Jinja2 library.
    """

    @classmethod
    def render_dir(cls,
                   source_dir: str,
                   target_dir: str,
                   overwrite: bool = False,
                   **render_args):
        """
        Renders the entire template directory to the given `target_dir`,
        using the provided rendering args, preserving the directory structure of
        the `source_dir`. Files that are not templates are copied as is.

        When `target_dir` does not exist, it will be created. If any destination
        file inside `target_dir` already exists, a FileExistsError will be
        raised when `overwrite` is False (default). Otherwise, files are
        overwritten.

        Args:
            source_dir (str): Path to the source directory,
                containing template files (and potentially static files to be
                copied without modification).
            target_dir (str): Path to the target directory.
            overwrite (bool): If True, the destination folder will first
                be deleted if it already exists.
                If False (default), a FileExistsError will be raised if
                any file already exists in the target directory.
            render_args: Keyword arguments to render the template files.

        Raises:
            FileExistsError: If the destination file already exists and
                `overwrite` is False.
        """
        if not source_dir:
            raise ValueError("Source directory not provided.")

        source_dir = pathlib.Path(source_dir)
        if not source_dir.is_dir():
            raise ValueError(f"Source directory {source_dir} does not exist.")

        if not target_dir:
            raise ValueError("Target directory not provided.")

        target_dir = pathlib.Path(target_dir)

        loader = jinja2.FileSystemLoader(source_dir)
        environment = jinja2.Environment(loader=loader,
                                         keep_trailing_newline=True,
                                         undefined=jinja2.StrictUndefined)

        template_dir_struct = helpers.get_dir_structure(source_dir)

        if target_dir.exists() and any(target_dir.iterdir()):
            if overwrite:
                shutil.rmtree(target_dir)
            else:
                msg = (f"Non-empty target directory {target_dir}.\n"
                       "Use `overwrite=True` to delete previous content.")
                raise FileExistsError(msg)

        for subdir, contents in template_dir_struct.items():
            target_subdir = target_dir / subdir
            target_subdir.mkdir(parents=True, exist_ok=True)
            source_subdir = source_dir / subdir

            # simply copy non-template files to destination
            for file in contents["files"]:
                shutil.copy(source_subdir / file, target_subdir)

            # render template files
            for file in contents["templates"]:
                target_path = target_subdir / helpers.strip_extension(file)
                template_name = (pathlib.Path(subdir) / file).as_posix()
                template = environment.get_template(template_name)
                stream = template.stream(**render_args)
                stream.dump(str(target_path))



"""DualSPHysics simulator module of the API."""

from typing import Literal, Optional, Union

from inductiva import simulators, tasks, types


class DualSPHysics(simulators.Simulator):
    """Class to invoke a generic DualSPHysics simulation on the API."""

    def __init__(self,
                 /,
                 version: Optional[str] = None,
                 use_dev: bool = False,
                 device: Literal["auto", "cpu", "gpu"] = "auto"):
        """Initialize the DualSPHysics simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
            device (str): Select between CPU or GPU for running the simulation.
                Default is "auto", which will auto-detect if the machine has a
                GPU and use if it available, otherwise use the CPU.
        """
        super().__init__(version=version, use_dev=use_dev, device=device)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "dualsphysics"

    def run(
        self,
        input_dir: Optional[str],
        shell_script: str,
        *,
        on: types.ComputationalResources,
        storage_dir: Optional[str] = "",
        resubmit_on_preemption: bool = False,
        remote_assets: Optional[Union[str, list[str]]] = None,
        project: Optional[str] = None,
        time_to_live: Optional[str] = None,
        vtk_to_obj: Optional[bool] = False,
        vtk_to_obj_vtk_dir: Optional[str] = None,
        vtk_to_obj_vtk_prefix: Optional[str] = "PartFluid_",
        vtk_to_obj_out_dir: Optional[str] = None,
        vtk_to_obj_particle_radius: Optional[float] = None,
        vtk_to_obj_smoothing_length: Optional[float] = 2.0,
        vtk_to_obj_cube_size: Optional[float] = 1.0,
        vtk_to_obj_surface_threshold: Optional[float] = 0.6,
        **kwargs,
    ) -> tasks.Task:
        """Executes a DualSPHysics simulation.

        Args:
            input_dir: Directory with simulation input files.
            shell_script: Path to the shell script to run the simulation.
            on: The computational resource to launch the simulation on.
            storage_dir: Directory for storing results.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            project: Name of the project to which the task will be
                assigned. If None, the task will be assigned to
                the default project. If the project does not exist, it will be
                created.
            time_to_live: Maximum allowed runtime for the task, specified as a
                string duration. Supports common time duration formats such as
                "10m", "2 hours", "1h30m", or "90s". The task will be
                automatically terminated if it exceeds this duration after
                starting.
            vtk_to_obj: Whether to convert the output VTK files to OBJ meshes
                using marching cubes.
            vtk_to_obj_vtk_dir: Directory containing VTK files to be converted.
            vtk_to_obj_out_dir: Directory where the generated OBJ files will be
                stored.
            vtk_to_obj_vtk_prefix: Prefix of the VTK files that will be
                converted.
                Default: PartFluid_ 
            vtk_to_obj_particle_radius: The particle radius of the input data.
            vtk_to_obj_smoothing_length: The smoothing length radius used for
                the SPH kernel, the kernel compact support radius will be twice
                the smoothing length (in multiplies of the particle radius).
                Default: 2.0
            vtk_to_obj_cube_size: The cube edge length used for marching cubes
                in multiplies of the particle radius, corresponds to the cell
                size of the implicit background grid.
                Default: 1.0
            vtk_to_obj_surface_threshold: The iso-surface threshold for the
                density, i.e. the normalized value of the reconstructed density
                level that indicates the fluid surface (in multiplies of the
                rest density).
                Default: 0.6
        Returns:
            tasks.Task: An object representing the simulation task.
        """

        self._input_files_exist(input_dir=input_dir,
                                remote_assets=remote_assets,
                                shell_script=shell_script)

        if vtk_to_obj and (vtk_to_obj_vtk_dir is None or
                           vtk_to_obj_particle_radius is None):
            raise ValueError("When using `vtk_to_obj=True`, "
                             "`vtk_to_obj_vtk_dir` and "
                             "`vtk_to_obj_particle_radius` need to be defined.")

        commands = [f"bash {shell_script}"]

        if vtk_to_obj:

            # If out_dir is not provided, will save in the same directory as
            # the vtk files
            if vtk_to_obj_out_dir is None:
                vtk_to_obj_out_dir = vtk_to_obj_vtk_dir

            commands.append(
                "splashsurf reconstruct "
                f"{vtk_to_obj_vtk_dir}/{vtk_to_obj_vtk_prefix}"
                "{}.vtk "
                f"-r={vtk_to_obj_particle_radius} "
                f"-l={vtk_to_obj_smoothing_length} "
                f"-c={vtk_to_obj_cube_size} "
                f"-t={vtk_to_obj_surface_threshold} "
                "--subdomain-grid=on --mesh-cleanup=on "
                "--mesh-smoothing-weights=on --mesh-smoothing-iters=25 "
                "--normals=on --normals-smoothing-iters=10 "
                f"-o {vtk_to_obj_out_dir}/{vtk_to_obj_vtk_prefix}_surface"
                "{}.obj")
        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           project=project,
                           time_to_live=time_to_live,
                           **kwargs)



"""
Uploads a Docker image as a converted Apptainer .sif to remote storage.
"""

import argparse
import os
import pathlib
import sys
import tempfile
from inductiva import storage, constants
from inductiva.client.exceptions import ApiException
from inductiva.utils.input_functions import user_confirmation_prompt
from .convert import convert_image


def extract_image_name(image_ref: str) -> str:
    """
    Extracts the base image name from a docker image reference.
    e.g.:
        docker://nginx:alpine → nginx
        nginx:alpine → nginx
        myorg/python → python
        python → python
    """
    # Strip docker:// if present
    image_ref = image_ref.removeprefix("docker://")

    # Split org/image and version tag
    name_part = image_ref.split("/")[-1]
    image_name = name_part.split(":")[0]
    return image_name


def upload_container(args):
    image_name = extract_image_name(args.image)
    default_folder = "my-containers"

    # Handle missing or partial output_path
    output_path = args.output_path

    if not output_path:
        output_path = f"{default_folder}/{image_name}.sif"
    else:
        output_path = os.path.normpath(output_path)

        # If it's just a filename (no folder), prepend default folder
        if os.sep not in output_path:
            output_path = os.path.join(default_folder, output_path)

        # Ensure .sif extension
        if not output_path.endswith(".sif"):
            output_path += ".sif"

    # Extract folder and filename from the now-final output_path
    folder_name = output_path.split(os.sep)[0]
    filename = os.path.basename(output_path)

    # ---------- pre-flight: does it already exist? ---------------------------
    try:
        contents = storage.listdir(folder_name, print_results=False)
    except ApiException as e:
        if e.status == 404:
            contents = []
        else:
            print(f"Error accessing remote folder '{folder_name}': {e}",
                  file=sys.stderr)
            return

    already_there = any(c["content_name"] == filename for c in contents)
    if already_there:
        if args.overwrite:
            print(f"'{output_path}' exists - will overwrite it "
                  "(because --overwrite).")
            try:
                storage.remove(f"{folder_name}/{filename}")
            except ApiException as e:
                print(f"Failed to delete old file: {e}", file=sys.stderr)
                return
        else:
            confirm = user_confirmation_prompt(
                [filename],
                "File already exists.",
                "File already exists.",
                "Overwrite the existing file?",
                is_all=False,
            )
            if not confirm:
                print("Operation cancelled.")
                return
            storage.remove(f"{folder_name}/{filename}")

    # Create temp folder to hold .sif
    try:
        with tempfile.TemporaryDirectory(dir=constants.TMP_DIR) as tmp_dir:
            sif_folder_path = pathlib.Path(tmp_dir) / folder_name
            os.makedirs(sif_folder_path, exist_ok=True)
            sif_file_path = os.path.join(sif_folder_path, filename)

            convert_args = argparse.Namespace(image=args.image,
                                              output=sif_file_path)

            print(f"Converting {args.image} -> {sif_file_path}...")
            if not convert_image(convert_args):
                print("❌ Conversion failed.")
                return

            print(
                f"Uploading '{sif_folder_path}' to remote dir '{folder_name}'.."
            )
            storage.upload(local_path=sif_folder_path, remote_dir=folder_name)

            # Print the remote path
            print("✅ Upload complete.")
            print("To use the container, instantiate it with:")
            print(f"\t > inductiva://{folder_name}/{filename}")
    except Exception as e:  # pylint: disable=broad-exception-caught
        print("❌ Failed to upload container")
        print(f"Error details: {str(e)}")
        return


def register(parser):
    """Register the upload-container command."""
    subparser = parser.add_parser(
        "upload",
        help="Convert a Docker image to a .sif and upload to remote storage.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    subparser.description = (
        "Converts a Docker image (from Docker Hub or local) into a .sif file "
        "using Apptainer,\n stores it temporarily in a folder, and uploads that"
        " folder to the system's remote storage.")

    subparser.add_argument(
        "image",
        type=str,
        help="Docker image reference (e.g., python:3.11-slim or docker://...).",
    )
    subparser.add_argument(
        "output_path",
        nargs="?",
        type=str,
        help=(
            "Optional output path for the .sif file, my-containers/nginx.sif.\n"
            "If omitted, defaults to my-containers/<image-name>.sif."),
    )
    subparser.add_argument(
        "-f",
        "--overwrite",
        action="store_true",
        help="Overwrite the file in remote storage without asking.")

    subparser.set_defaults(func=upload_container)
